#include <xc.h>
#include <stdio.h>
#include "SDlib16_delay.h"
#include "regList.c"

#define FREQUENCY 905700000     // Channel 17

// DEVCFG3
#pragma config USERID = 0xFFFF          // Enter Hexadecimal value (Enter Hexadecimal value)
#pragma config PMDL1WAY = ON            // Peripheral Module Disable Configuration (Allow only one reconfiguration)
#pragma config IOL1WAY = ON             // Peripheral Pin Select Configuration (Allow only one reconfiguration)

// DEVCFG2
#pragma config FPLLIDIV = DIV_2         // PLL Input Divider (2x Divider)
#pragma config FPLLMUL = MUL_20         // PLL Multiplier (20x Multiplier)
#pragma config FPLLODIV = DIV_2         // System PLL Output Clock Divider (PLL Divide by 2)

// DEVCFG1
#pragma config FNOSC = FRCPLL           // Oscillator Selection Bits (Fast RC Osc with PLL)
#pragma config FSOSCEN = OFF            // Secondary Oscillator Enable (Disabled)
#pragma config IESO = ON                // Internal/External Switch Over (Enabled)
#pragma config POSCMOD = OFF            // Primary Oscillator Configuration (Primary osc disabled)
#pragma config OSCIOFNC = OFF           // CLKO Output Signal Active on the OSCO Pin (Disabled)
#pragma config FPBDIV = DIV_4           // Peripheral Clock Divisor (Pb_Clk is Sys_Clk/4)
#pragma config FCKSM = CSECME           // Clock Switching and Monitor Selection (Clock Switch Enable, FSCM Enabled)
#pragma config WDTPS = PS4096           // Watchdog Timer Postscaler (1:4096)
#pragma config WINDIS = OFF             // Watchdog Timer Window Enable (Watchdog Timer is in Non-Window Mode)
#pragma config FWDTEN = OFF             // Watchdog Timer Enable (WDT Disabled (SWDTEN Bit Controls))
#pragma config FWDTWINSZ = WINSZ_25     // Watchdog Timer Window Size (Window Size is 25%)

// DEVCFG0
#pragma config JTAGEN = OFF             // JTAG Enable (JTAG Disabled)
#pragma config ICESEL = ICS_PGx1        // ICE/ICD Comm Channel Select (Communicate on PGEC1/PGED1)
#pragma config PWP = OFF                // Program Flash Write Protect (Disable)
#pragma config BWP = OFF                // Boot Flash Write Protect bit (Protection Disabled)
#pragma config CP = OFF                 // Code Protect (Protection Disabled)



char uart1_init(unsigned int baud) {
    U1RXR = 0b0001; // RB6 is U1RX
    RPB7R = 0b0001; // RB7 is U1TX
    U1MODE = 0x00000088;
    U1BRG = 10000000ul / (4*baud) - 1;
    U1STAbits.UTXEN = 1;
    U1STAbits.URXEN = 1;
    U1MODEbits.ON = 1;
    return 0;
}

char uart2_init(unsigned int baud) {
    U2RXR = 0b0100; // RB8 is U2RX
    RPB9R = 0b0010; // RB9 is U2TX
    U2MODE = 0x00000088;
    U2BRG = 10000000ul / (4*baud) - 1;
    U2STAbits.UTXEN = 1;
    U2STAbits.URXEN = 1;
    U2MODEbits.ON = 1;
    return 0;
}

char uart1_write(char tx) {
    while(U1STAbits.UTXBF); // Block while transmit buffer is full
    U1TXREG = tx;
    return tx;
}

char uart2_write(char tx) {
    while(U2STAbits.UTXBF); // Block while transmit buffer is full
    U2TXREG = tx;
    return tx;
}

char write_uart(char moduleNum, char tx) {
    switch(moduleNum) {
	case 1:
	    return uart1_write(tx);
	    break;
	case 2:
	    return uart2_write(tx);
	    break;
	default:
	    return -8; // Return -8 if moduleNum is not valid
    }
}

char uart1_read() {
    if(U1STAbits.URXDA)
        return U1RXREG;
    return -7; // Return -7 if receive buffer is empty
}

char uart2_read() {
    if(U2STAbits.URXDA)
        return U2RXREG;
    return -7; // Return -7 if receive buffer is empty
}

char read_uart(char moduleNum) {
    switch (moduleNum){
	case 1:
	    return uart1_read();
	    break;
	case 2:
	    return uart2_read();
	    break;
	default:
	    return -8; //Return -8 if moduleNum is not valid
    }
}

char uart1_wait_for_read() {
    while(!U1STAbits.URXDA);
    return U1RXREG;
}

char uart2_wait_for_read() {
    while(!U2STAbits.URXDA);
    return U2RXREG;
}

char wait_for_read_uart(char moduleNum) {
    switch(moduleNum) {
	case 1:
	    return uart1_wait_for_read();
	    break;
	case 2:
	    return uart2_wait_for_read();
	    break;
	default:
	    return -8;
    }
}

char spi_init(unsigned int sck) {
//    SPI1CON = 0x100120AD; // 8-bit words, not Framed mode, with SS
    SPI1CONbits.ON = 0;
    SPI1CONbits.CKE = 1;
    SPI1CONbits.CKP = 0;
    SPI1CONbits.MSTEN = 1;
    SPI1CONbits.ENHBUF = 0;
    SPI1BRG = 10000000ul / (2*sck) - 1;
    RPB13R = 0x3; // RB13 is SDO1
    SDI1R = 0x3; // RB11 is SDI1
    // RB15 is SS1 (active low)
    SPI1CONbits.ON = 1;
    return 0;
}

char read_spi() {
    while(SPI1STATbits.SPIRBE); // Block while waiting for receive
    delay_us(5);
    return SPI1BUF;
    return -7; // If receive buffer is empty, return -7 (illegal receive value)
}

char write_spi(char tx) {
    while(SPI1STATbits.SPITBF) { // While transmit buffer is full, wait until space opens
	delay_us(1);
    }
    delay_us(5);
    return SPI1BUF = tx;
}

void write_reg_spi(char addr, char val) {
    LATBbits.LATB15 = 0;
    delay_us(5);
    write_spi(0x80 | addr);
    read_spi();
    write_spi(val);
    read_spi();
    delay_us(5);
    LATBbits.LATB15 = 1;
    delay_us(20);
}

char read_reg_spi(char addr) {
    LATBbits.LATB15 = 0;
    delay_us(5);
    write_spi(addr);
    read_spi();
    write_spi(0x00);
    char val = read_spi();
    delay_us(5);
    LATBbits.LATB15 = 1;
    delay_us(20);
    return val;
}

char write_reg_burst(char addr, char* vals, char len) {
    char i;
    LATBbits.LATB15 = 0;
    delay_us(5);
    write_spi(0x80 | addr);
    for(i = 0; i < len; i++) {
	delay_us(15);
	write_spi(vals[i]);
	read_spi();
    }
    LATBbits.LATB15 = 1;
    return len;
}

char* read_reg_burst(char addr, char* vals, char len) {
    LATBbits.LATB15 = 0;
    delay_us(5);
    write_spi(addr);
    read_spi();
    for(char i = 0; i < len; i++) {
	write_spi(0x00);
	vals[i] = read_spi();
    }
    delay_us(5);
    LATBbits.LATB15 = 1;
}

int radio_config() {
    uint64_t frf = ((uint64_t)FREQUENCY << 19) / 32000000;

    uint8_t frf_msb = (uint8_t)(frf >> 16);
    uint8_t frf_mid = (uint8_t)(frf >> 8);
    uint8_t frf_lsb = (uint8_t)(frf >> 0);
    
    
    read_reg_spi(REG_VERSION);
    write_reg_spi(REG_OP_MODE, (MODE_LONG_RANGE | MODE_SLEEP)); //Set RegOpMode to sleep
    delay_us(50);
    read_reg_spi(REG_OP_MODE);
    delay_us(50);
//    write_reg_spi(REG_FRF_MSB, 0xE4);
//    write_reg_spi(REG_FRF_MID, 0xC0);
//    write_reg_spi(REG_FRF_LSB, 0x00);
    write_reg_spi(REG_FRF_MSB, frf_msb);
    write_reg_spi(REG_FRF_MID, frf_mid);
    write_reg_spi(REG_FRF_LSB, frf_lsb);
    write_reg_spi(REG_FIFO_TX_BASE_ADDR, 0x00);
    write_reg_spi(REG_FIFO_RX_BASE_ADDR, 0x00);
    short lnaGain = read_reg_spi(REG_LNA) & 0xE0;
    write_reg_spi(REG_LNA, (lnaGain | 0x03));
    write_reg_spi(REG_MODEM_CONFIG_3, 0x04);
    write_reg_spi(REG_PA_DAC, 0x84);
    write_reg_spi(REG_OCP, 0x2B);
    write_reg_spi(REG_PA_CONFIG, 0x8F);
    write_reg_spi(REG_RX_TIMEOUT, 0xFF);
    write_reg_spi(REG_OP_MODE, (MODE_LONG_RANGE | MODE_STDBY));
    delay_ms(5);
    return 0;
}

void rf_block_while_tx() {
    while(!(read_reg_spi(REG_IRQ_FLAGS)) & MASK_TX_DONE)
	delay_us(100);
    write_reg_spi(REG_IRQ_FLAGS, MASK_TX_DONE);
    return;
}

char begin_packet() {
    write_reg_spi(REG_OP_MODE, (MODE_LONG_RANGE | MODE_STDBY));
    write_reg_spi(REG_MODEM_CONFIG_1, 0x72);
    write_reg_spi(REG_FIFO_ADDR_PTR, 0x00);
    write_reg_spi(REG_PAYLOAD_LENGTH, 0x00);
    return 0;
}

void add_data_to_packet(char data) {
    write_reg_spi(REG_FIFO, data);
//    *ptrPayloadLen++; // Pointer is used so the payload length is automatically incremented by the function
//    return *ptrPayloadLen;
}
    
char end_packet(char payloadLen) {
    write_reg_spi(REG_PAYLOAD_LENGTH, payloadLen);
    write_reg_spi(REG_OP_MODE, (MODE_LONG_RANGE | MODE_TX));
    return payloadLen;
}

char send_packet(char* payload, char payloadLen) {
    int i = 0;
    begin_packet();
//    write_reg_burst(REG_FIFO, payload, payloadLen);
    for(i = 0; i < payloadLen; i++) {
	add_data_to_packet(payload[i]);
    }
    return end_packet(payloadLen);
}

char rf_block_until_rx() {
    while(!(read_reg_spi(REG_IRQ_FLAGS) & (MASK_RX_TIMEOUT | MASK_RX_DONE)))
	delay_us(50);
    char irqFlags = (read_reg_spi(REG_IRQ_FLAGS));
    write_reg_spi(REG_IRQ_FLAGS, (MASK_RX_TIMEOUT | MASK_RX_DONE | MASK_PAYLOAD_CRC_ERROR | MASK_VALID_HEADER));
    return irqFlags;
}

int main() {
    unsigned char payloadLen = 0;
    unsigned char payload[64];
    
    // RUN_TIME DEBUGGING VARIABLES
    
    char tx = 1; // If tx == 1, transmitter mode enabled
    char uart_debug = 2;
    
    delay_ms(500);
    ANSELA = 0;
    ANSELB = 0;
    TRISBbits.TRISB15 = 0;
    LATBbits.LATB15 = 1;
    spi_init(100000);
    uart1_init(57600);
    uart2_init(57600);
    unsigned char row = '0';
    unsigned char col = 'A';
    unsigned char i = 0;
    radio_config();
    
    delay_ms(250);
    uart2_write('B');
    uart2_write('o');
    uart2_write('o');
    uart2_write('t');
    uart2_write('i');
    uart2_write('n');
    uart2_write('g');
    uart2_write('\r');
    uart2_write('\n');
    delay_ms(25);
    
    if(tx) {
	payload[0] = 'S';
	payload[1] = 't';
	payload[2] = 'a';
	payload[3] = 'r';
	payload[4] = 't';
	payload[5] = 'i';
	payload[6] = 'n';
	payload[7] = 'g';
	payload[8] = '\r';
	payload[9] = '\n';
	payloadLen = 10;
	send_packet(payload, payloadLen);
	rf_block_while_tx();
    }
    
    while(tx) {
	payloadLen = 0;
//	payload[0] = col++;
//	payload[1] = row++;
//	payload[2] = '\r';
//	payload[3] = '\n';
//	payloadLen = 4;
//	if(col > 'Z')
//	    col = 'A';
//	if(row > '9')
//	    row = '0';
	do {
	    payload[payloadLen] = write_uart(uart_debug, wait_for_read_uart(1));
	    payloadLen++;
	} while(payload[payloadLen-1] != '\n'
	    && payload[payloadLen-1] != '\r'
	    && payloadLen < 64);
	if(payload[payloadLen-1] == '\r' && payloadLen < 63) {
	    payload[payloadLen] = '\n';
	    write_uart(uart_debug, '\n');
	    payloadLen++;
	}
	
	
//	payload[0] = 'H';
//	payload[1] = 'e';
//	payload[2] = 'l';
//	payload[3] = 'l';
//	payload[4] = 'o';
//	payload[5] = '\n';
//	payloadLen = 6;
	send_packet(payload, payloadLen);
	rf_block_while_tx();
	for(i = 0; i < payloadLen; i++) {
	    write_uart(uart_debug, payload[i]);
	}
	delay_ms(250);	   
    }
    
    while(!tx) {
	write_reg_spi(REG_OP_MODE, (MODE_LONG_RANGE | MODE_RX_SINGLE));
	write_reg_spi(REG_FIFO_ADDR_PTR, 0x00);
	unsigned char irqFlags = rf_block_until_rx();
	switch(irqFlags) {
	    case (MASK_RX_DONE | MASK_VALID_HEADER):
		payloadLen = read_reg_spi(REG_RX_NB_BYTES);
		read_reg_burst(REG_FIFO, payload, payloadLen);
		write_reg_spi(REG_OP_MODE, (MODE_LONG_RANGE | MODE_STDBY));
//		write_uart(uart_debug, '\r');
//		write_uart(uart_debug, '\n');
		for(i = 0; i < payloadLen; i++)
		    write_uart(uart_debug, payload[i]);
		break;
	    case (MASK_RX_DONE | MASK_PAYLOAD_CRC_ERROR | MASK_VALID_HEADER):
		write_reg_spi(REG_OP_MODE, (MODE_LONG_RANGE | MODE_STDBY));
		break;
	    case (MASK_RX_TIMEOUT):
		write_reg_spi(REG_OP_MODE, (MODE_LONG_RANGE | MODE_STDBY));
		break;
	    default:
		write_reg_spi(REG_OP_MODE, (MODE_LONG_RANGE | MODE_STDBY));
	}
	delay_us(25);
    }
    
    return (EXIT_SUCCESS);
}

